-- vi: et sw=4 ts=4
-- hsv2
---@module "hsv2.config"
local M = {}
local L = {}

local defaults = require("hsv2.config.defaults")
local utils    = require("hsv2.utils")

---Get config key value.
---@param  path?            path|path_raw
---@param  return_root?     boolean             Return a table with full config table hierarchy.
---@param  dont_copy_ref?   boolean             Don't copy value even if it's a reference type.
---@return any
function M.get(path, return_root, dont_copy_ref)
    local value = utils.tables.get(L.config(), path, defaults, return_root)

    return dont_copy_ref
        and value
        or utils.generics.ensure_copy(value)
end

---Set a config key value.
---@param path  path|path_raw
---@param value any
function M.set(path, value)
    local function handler(p, v)
        L.on_config_update(p, v)
    end

    if type(value) == "table" then
        utils.tables.rmap(utils.tables.nest_value(path, value), function(_, p, _, v)
            handler(p, v)
        end)
    else
        handler(path, value)
    end
end

---@alias config_listener_fn fun(c: table<string, any>, p: path, v: any)
---Config listener table path -> priority group -> listeners
---@type  table<string, table<integer, config_listener[]>>
L.listeners = {}


---@class config_listener
---@field  fn        config_listener_fn
---@field  name?     string  descriptive name
---@field  priority  integer [1-100]
---@param  path      path
---@param  listener  config_listener_fn
---@return config_listener[]
function L.get_listener(path, listener)
    utils.asserts.type(listener, "function", "listener")

    local listeners = L.get_listeners(path)
    for _, v in ipairs(listeners) do
        if v.fn == listener then return v end
    end
end

---Get config listeners for a path, ordered by priority
---@param  path path
---@return config_listener[]
function L.get_listeners(path)
    ---@type config_listener[]
    local path_listeners = {}
    for p, prio_group in pairs(L.listeners) do
        if utils.paths.compare(path, p) then
            for i = 1, 100 do
                if prio_group[i] then
                    local group = vim.tbl_extend("force", {}, prio_group[i])
                    for _, listener in ipairs(group) do
                        listener.priority = i
                    end
                    vim.list_extend(path_listeners, group)
                end
            end
        end
    end

    return path_listeners
end

---Register a config listener
---@param path     path|path_raw
---@param listener config_listener
---@param name?    string
---@param priority integer [1-100]
---@return boolean true  listener registered.
---@return boolean false same listener with same priority already registered.
function M.register_listener(path, listener, name, priority)
    utils.asserts.type(path, { "string", "number", "table" }, "path")
    utils.asserts.type(listener, "function", "listener")
    utils.asserts.type(name, "string", "name", true)
    utils.asserts.type(priority, "number", "priority", true)

    listener = { fn = listener, name = name}

    local current_listener = L.get_listener(path, listener.fn)
    if current_listener then
        if current_listener.priority == priority then
            return false
        end
        M.unregister_listener(path, current_listener.fn, current_listener.priority)
    end

    path                        = utils.paths.to_string(path)
    priority                    = priority and priority <= 100 and priority > 0 and priority or L.listener_def_prio
    name                        = name                                          or "no name"
    L.listeners[path]           = L.listeners[path]                             or {}
    L.listeners[path][priority] = L.listeners[path][priority]                   or {}

    table.insert(L.listeners[path][priority], listener)

    return true
end

---Unregister a config listener.
---@param  path             path|path_raw
---@param  listener         config_listener
---@param  priority         integer         [1-100]
---@return config_listener? removed         removed listener
function M.unregister_listener(path, listener, priority)
    utils.asserts.type(listener, "function", "listener")
    utils.asserts.type(priority, "number", "priority")

    path     = utils.paths.to_string(path)
    priority = priority <= 100 and priority > 0 and priority or L.listener_def_prio

    for i, v in ipairs(L.listeners[path][priority]) do
        if v.fn == listener then
            return table.remove(L.listeners[path][priority], i)
        end
    end
end

L.config = (function()
    ---@class hsv2_config
    ---@field log? LogConfig
    local config = {}

    return function()
        config = vim.tbl_deep_extend("keep", config, defaults())
        return config
    end
end)()

---Called on each new config table leaf nodes.
---@param  path     path|path_raw
---@param  value    any
---@return boolean, table [1] if a config_listener gaves an error returns false, true otherwise and a list of status of each config_listener call
function L.on_config_update(path, value)
    local error     = false
    local listeners = L.get_listeners(path)
    local status    = {}
    for _, l in ipairs(listeners) do
        local s = l.fn(L.config(), path, value)
        error = error or not s
    end

    return not error, status
end

local function noop(...)
    return ...
end
---@alias env_preproc_fn fun(env_value: string):any function used to preprocess env values
---@type  table<string, table<path_raw, env_preproc_fn>> map env variable name to config table path and a preprocessing function
L.map_env_conf = {
    HSV2_LOG_LEVEL = { "log.level",  noop},
}

---Load configuration from environment
---@return hsv2_config conf
function M.from_env()
    local conf = {}
    for env, pair in pairs(L.map_env_conf) do
        utils.tables.set(conf, pair[1], pair[2](vim.env[env]))
    end

    vim.pretty_print({env=conf})
    return conf
end
--[[
function M.config_from_env()
    local one_of = require("hsv2.utils").is_one_of_values
    local env_config = {}

    local log_level = vim.env.HSV2_LOG_LEVEL
    if one_of(type(log_level), { "string", "number" }) then
        u().tables.deprecated.table_append(env_config, {
            log = {
                level = log_level,
            },
        })
    end

    return env_config
end

function M.defaults(path)
    return u().tables.get(require('hsv2.config.default'), path)
end

---Return current config discarding default values.
---@param  path  path|nil Path to get from configuration table.
---@return table conf     The current configuration without defaults.
function M.defaults_free(path)
    local conf = u().tables.copy(M.config())
    u().tables.rmap(M.defaults(), function(_, path, _, default)

L.defaults = (function()
    local defaults = require("hsv2.config.defaults")

    return function()
        return defaults
    end
end)()
        local current = u().tables.get(conf, path)
        if current == default then
            vim.pretty_print("nil", path)
            u().tables.set(conf, path)
        end
    end, path)
    return conf
end

function M.on_config_log_change(config_override)
    M.log_level_set(config_override.level)
end

function M.log_level_set(level)
    level = M.log_level_name(level)
    assert(u().tables.one_of_keys(level, M.config("enum.log_levels")),
        "Log level " .. tostring(level) .. " is invalid. Possible values: "
        .. require("hsv2.utils").list_keys_str(M.config("enum.log_levels")) .. "\n" .. debug.traceback())

    local previous_level = M.log_level_name(M.log_level()) or "none"
    config.log.level = level
    require("hsv2.utils.cmd").echo_info({
        "Set log level to ",
        { level, "Keyword" },
        ", was ",
        { previous_level, "Keyword" },
        ".",
    })
end

function M.log_level(level)
    if u().table.one_of_values(type(level), {
        "string",
        "number",
    }) then
        return level <= M.log_level_number(M.config("log.level"))
    end

    return M.config("log.level")
end

function M.log_level_number(level)
    if type(level) == "number" then
        return level
    end
    assert(u().tables.one_of_keys(M.log_level_name(level), M.config("enum.log_levels")))

function M.is_one_of_keys(key, map)
    assert(type(map) == "table")
    for k in pairs(map) do
        if key == k then
            return true
        end
    end
    return false
end
    return M.config({
        "enum",
        "log_levels",
        level,
    })
end

function M.log_level_name(level)
    assert(level ~= nil, "invalid parameter " .. vim.inspect(level))
    if type(level) == "string" then
        return level
    end

    if type(level) == "number" then
        for lvlname, code in pairs(M.config("enum.log_levels")) do
            if code == level then
                return lvlname
            end
        end
    end
end

function M.requirements()
    local append = require("hsv2.utils").table_append
    local req = M.requirements_hsv2()
    append(req, M.requirements_plugins())
    return req
end

function M.requirements_hsv2()
    local requirements = {}
    for name, req in pairs(M.hsv2().requirements) do
        req.by = "hsv2"
        requirements[name] = req
    end
    return requirements
end

function M.requirements_plugin(plugin)
    assert(type(plugin) == "string", "Parameter plugin must be a string.")
    local requirements = M.plugin(plugin).requirements
    for req, _ in pairs(requirements) do
        requirements[req].by = plugin
    end
    return requirements
end

function M.requirements_plugins(plugins)
    local param_type = type(plugins)
    assert(require("hsv2.utils").is_one_of_values(param_type, {
        "string",
        "table",
        "nil",
    }), "Parameter plugins must be either a string, a list of string or nil.")
    plugins = param_type == "table" and plugins or { plugins }
    local requirements = {}
    local append = require("hsv2.utils").table_append

    if next(plugins) == nil then
        for plugin, conf in pairs(M.plugins()) do
            if conf.enabled and type(conf.requirements) == "table" then
                table.insert(plugins, plugin)
            end
        end
    end
    for _, plugin in pairs(plugins) do
        append(requirements, M.requirements_plugin(plugin))
    end

    return requirements
end

function L.set_defaults(t1, defaults)
    local set = function(k, v)
        if type(t1[k]) == type(v) == 'table' then
            t1[k] = M.set_defaults(t1[k], v)
        elseif not t1[k] then
            t1[k] = v
        end
    end
    t1 = type(t1) == 'table' and t1 or {}
    if type(defaults) == 'table' then
        M.set_defaults(defaults)
    else
        defaults = M.defaults()
    end
    for k, v in pairs(defaults) do
        set(k, v)
    end
    return t1
end

function L.expand_config(conf)
    L.expand_packer_config(conf)
    return conf
end

function L.expand_packer_config(conf)
    if conf.hsv2.use_icons then
        conf.packer.display.working_sym = conf.packer.display.working_sym or ''
        conf.packer.display.removed_sym = conf.packer.display.removed_sym or ''
        conf.packer.display.header_sym = conf.packer.display.header_sym or ''
    end
    -- if M.debug() then
    --     config.packer.log = { level = 'debug' }
    -- elseif M.verbose() then
    --     config.packer.log = { level = 'info' }
    -- end
    return conf
end
--]]

return M
