-- vi: et ts=4 sw=4:
-- hsv-2
---@module "hsv2.utils.log"
local M = {}
local L = {}

---@class LogConfig @Logging configuration
---@field level log_level
local log_config = {
    level = "notice",
}

---Init log module
---@param config hsv2_config
function M.init(config)
    ---@type config_listener
    function M.config_listener(_, path, value)
        vim.pretty_print({path=path,value=value})
        local paths = require("hsv2.utils.paths")
        if paths.compare("log.level", path) == true then
            local level = L.level_get(value)
            if log_config.level ~= level then
                M.log({
                    { "log level changed: before '" },
                    { log_config.level, "Keyword" },
                    { "' - now '" },
                    { level, "Keyword" },
                    { "'" }
                }, "notice")
                log_config.level = level
            end
        end
    end

    ---Log message at given log level
    ---@param msg    echo_str[]|string
    ---@param level? log_level
    function M.log(msg, level)
        msg = type(msg) == "string" and { { msg } } or msg
        local io = require("hsv2.utils.io")
        if L.level_cmp(level, log_config.level) <= 0 then
            local msg_with_prefix = vim.list_extend(L.msg_prefix(level), msg)
            io.echo(msg_with_prefix, true)
        end
    end

    require("hsv2.utils.tables").rmap(config, M.config_listener, { "log" })
    config.log = log_config

    return M
end

---Compare log level priorities
---@param  l1      log_level
---@param  l2      log_level
---@return integer positive if l1 is greater than l2
---@return integer zero     if l1 and l2 are equals
---@return integer negative if l1 is lesser than l2
function L.level_cmp(l1, l2)
    local levels = require("hsv2.enums").log_level
    l1, l2 = levels[L.level_get(l1)], levels[L.level_get(l2)]
    return l1 - l2
end

---Takes a log level name or priority and return the level name if it exists or
---defaults to "notice".
---@param  level?    log_level|integer
---@return log_level
---|> "notice"             if level is not a valid log level
---|  "debug"              if level is "all"
---@return log_level level if level is a valid log level.
function L.level_get(level)
    if not level then return "notice" end
    if level == "all" then return "debug" end

    for name, prio in pairs(require("hsv2.enums").log_level) do
        if level == name or level == prio then
            return name
        end
    end

    M.log({
        "invalid log level '",
        { tostring(level), "Keyword" },
        "' default to '",
        { "notice", "Keyword" },
        "'"
    }, "debug")

    return "notice"
end

---Get log message prefix string
---@param level?      log_level
---@return echo_str[] prefix
function L.msg_prefix(level)
    level = L.level_get(level)
    local prefix = {
        { "HSV2:", "Title" },
        { level },
        { ": " },
    }
    prefix[2][2] = level == "err" and "Error" or level == "warn" and "Warning"
        or "Normal"

    return prefix
end

return M
