-- vi: et sw=4 ts=4:
-- hsv2
---@module "hsv2.utils.tables"
local M = {}
local L = {}

---Deep copy a table.
---@param  t1    table
---@return table t2    Copy of t1
function M.copy(t1)
    local asserts = require("hsv2.utils.asserts")
    asserts.type(t1, "table", "t1")

    local t2 = {}
    M.rmap(t1, function(_, path, _, value)
        L.set(t2, path, value)
    end)

    return t2
end

---Set default values when there's no current ones.
---@param  t1       table
---@param  path?    path
---@param  defaults table
---@return table    t1
function M.fill_defaults(t1, path, defaults)
    M.rmap(defaults, function(_, rmap_path, _, default)
        local current, status = L.get(t1, rmap_path)
        if (current == nil or not status) and default ~= nil then
            L.set(t1, rmap_path, default)
        end
    end, path)
    return t1
end

---Get value from a table with optional default one.
---@param  t1           table   Table to access.
---@param  path?        path    Table path to access.
---@param  defaults?    any
---@param  table_root?  boolean
---@return any          value?  Retuns the value at given path if table_root is false.
---@return table        res     Returns the value inside the full table hierarchy if table_root is true.
function M.get(t1, path, defaults, table_root)
    local a = require("hsv2.utils.asserts")
    a.type(t1, "table", "t1")
    a.type(table_root, "boolean", "table_root", true)

    local paths = require("hsv2.utils.paths")
    path = paths.parse(path)

    if type(defaults) == "table" and next(defaults) then
        M.fill_defaults(t1, path, defaults)
    end

    if not path then
        return t1, true
    end

    local current_value, status, msg = L.get(t1, path)
    local default_value = defaults
    if type(defaults) == "table" then
        default_value = L.get(defaults, path)
    end

    if not status and defaults == nil then
        return table_root and t1 or nil, true, msg
    elseif current_value == nil then
        L.set(t1, path, default_value)
        current_value = default_value
        msg = default_value ~= nil and "default value used"
            or not status and msg
            or nil
    end

    return table_root and t1 or current_value, true, msg
end

---Join various list into a new one.
---@param  ...   table
---@return table
function M.join_list(...)
    local res = {}
    for i = 1, select("#", ...) do
        local list = select(i, ...)
        require("hsv2.utils.asserts").type(list, "table", tostring(i))
        for _, item in ipairs(list) do
            table.insert(res, item)
        end
    end

    return res
end

---For (".", true) returns true.
---For ("foo.bar.baz", { 1 }) returns { foo = { bar = { baz = { 1 } }} }.
---@param  path         path|path_raw
---@param  value        any
---@return table|any
function M.nest_value(path, value)
    path = require("hsv2.utils.paths").parse(path)
    if not path then return value end

    local t1 = {}
    L.set(t1, path, value)

    return t1
end

function M.one_of_keys(key, map)
    assert(type(map) == "table")
    for k in pairs(map) do
        if key == k then
            return true
        end
    end
    return false
end

---Check if value is one of values
---@param  value    any     Value to check.
---@param  expected any     A lists of possible values.
---@return boolean
---@nodiscard
function M.one_of_values(value, expected)
    for _, v in pairs(expected) do
        if value == v then
            return true
        end
    end
    return false
end

---Apply a function to each keys of a table, optionaly below a given path.
---@alias  tables_rmap_cb fun(t1: table, cb: function, path: path)
---@param  t1       table
---@param  callback tables_rmap_cb
---@param  path     path?
function M.rmap(t1, callback, path)
    local asserts = require("hsv2.utils.asserts")
    asserts.type(t1, "table", "t1")
    asserts.type(callback, "function", "callback")
    asserts.type(path, "table", "path", true)

    path = path or {}

    local parent = L.get(t1, { unpack(path, 1, #path - 1) })
    local value, status = L.get(t1, path)
    if type(parent) ~= "table" or not status then
        return nil, false, "invalid path"
    end

    if type(value) == "table" then
        local child_key_idx = #path + 1
        for k in pairs(value) do
            local subpath = { unpack(path) }
            subpath[child_key_idx] = k
            M.rmap(t1, callback, subpath)
        end
    else
        callback(t1, path, parent, value)
    end

    return t1, true
end

---Set value in a table at a given path.
---@param  t1            table  Table to write to.
---@param  path          path   Path.
---@param  value         any?   Value to set.
---@return table         t1
function M.set(t1, path, value, return_value_parent)
    local asserts = require("hsv2.utils.asserts")
    asserts.type(t1, "table", "t1")
    asserts.type(path, { "table", "number", "string" }, "path")
    path = require("hsv2.utils.paths").parse(path)
    assert(next(path), "path cannot be empty")

    local current, status, msg = L.get(t1, path)
    if not status and current ~= nil then
        return nil, false, msg
    end
    L.set(t1, path, value)

    return not return_value_parent and t1 or L.get(t1, { unpack(path, 1, #path - 1) })
end

---TODO
---@param t1        table
---@param schema    table
---@return boolean
function M.validate(t1, schema)
    return true
end

---Access a value in a table with a list of successive keys to use.
---@param  t1   table
---@param  path path
---@return any  value?
function L.get(t1, path)
    require("hsv2.utils.asserts").type(t1, "table", "t1")
    if not path then
        return t1
    end

    local value = t1
    for i, key in ipairs(path) do
        if type(value) ~= "table" then
            return nil, false, "path component '" .. tostring(path[i - 1])
                .. "' is not a table.", i - 1
        end
        value = value[key]
    end
    return value, true
end

---Set a value in a table.
---@param  t1       table
---@param  path     path
---@param  value    any
---@return table    t1
function L.set(t1, path, value)
    local asserts = require("hsv2.utils.asserts")
    asserts.type(t1, "table", "t1")
    asserts.type(path, "table", "path")
    assert(next(path))

    local node = t1
    for i, key in ipairs(path) do
        if i == #path then
            node[key] = value
        elseif type(node[key]) ~= "table" then
            node[key] = {}
        end
        node = node[key]
    end

    return t1
end

return M, L
