-- vi: et sw=4 ts=4:
-- hsv-2
local M = {
    sample = {},
    tests = {},
    fn = {},
}

M.sample.key_mapping = {
    a = {
        name = "test",
        b = {
            rhs = "<CMD>echo 'hello'<CR>",
            name = "test",
            opts = {
                silent = false,
            },
        },
        c = {
            name = "visual test",
            mode = "v",
            d = {
                name = "visual",
                mode = "v",
                rhs = "<CMD>echo 'visual'<CR>"
            }
        }
    },
}

function M.tests.str_split()
    local split = require("hsv2.utils.str").split
    for _, str in ipairs({
        "",
        ".",
        ".....",
        "foo",
        "foo.",
        ".foo",
        ".foo.",
        "foo.bar",
        "foo.bar.baz",
        "...foo...bar..baz...",
    }) do
        vim.pretty_print({
            str = str,
            keep = split(str, "%.", true),
            strip = split(str, "%.", false),
        })
    end
    for _, str in ipairs({
        "",
        ".",
        ",,;,a;.,a;.,,;",
        "fooa,;.",
        "foo.,a;bar",
        "foo.,;bar.,,,a;baz",
        ".,;.,;.,;foo.,;.,,,;.,,;bar..baz.,;.,;",
    }) do
        vim.pretty_print({
            str = str,
            keep = split(str, "%.,+a?;", true),
            strip = split(str, "%.,+a;", false),
        })
    end
end

function M.tests.tables(start_i, end_i)
    local nesting = { { { a = {} } } }
    local t1 = {
        bool = true,
        foo = {},
        sub1 = {
            a = 1,
            sub2 = {
                a = 2,
                sub3 = {
                    a = 3,
                    nesting = nesting,
                },
            },
        },
    }
    local defaults = {
        bool = false,
        foo = {
            bar = {
                baz = "foo.bar.baz"
            }
        },
        some_defaults = {
            defaults = {},
            def_a = 3.14,
        }
    }
    local tbl = require("hsv2.utils.tables")
    local function check_type(path, expected)
        return function(t)
            local value = tbl.get(t, path, defaults)
            return expected == type(value), value
        end
    end

    local function check_empty_tbl(path, negate)
        return function(t)
            local node = tbl.get(t, path, defaults)
            local empty = next(node) == nil
            if negate then
                return not empty, node
            end
            return empty, node
        end
    end

    local function check_value(path, expected)
        return function(t)
            local res_get = tbl.get(t, path, defaults)
            return res_get == expected and type(res_get) == type(expected),
                res_get
        end
    end

    local tests = {
        {
            expect = {
                [".bool == true"] = check_value("bool", true),
                [".foo is a table"] = check_type("foo", "table"),
                [".foo is not empty"] = check_empty_tbl("foo", true),
                [".foo.bar.baz == 'foo.bar.baz'"] = check_value("foo.bar.baz", "foo.bar.baz"),
                [".some_defaults.def_a == 3.14"] = check_value("some_defaults.def_a", 3.14),
                [".some_defaults.defaults is empty"] = check_empty_tbl("some_defaults.defaults"),
            },
        },
        {
            pre = function(t)
                tbl.set(t, "false_bool", false)
                tbl.set(t, "foo.bar.bazz", {})
                tbl.set(t, "sub1.a.overrided.nested", nesting)
            end,
            expect = {
                [".false_bool == false"] = check_value("false_bool", false),
                [".foo.bar.bazz is a table"] = check_type("foo.bar.bazz", "table"),
                [".sub1.a.overrided.nested is a table"] = check_type("sub1.a.overrided.nested", "table"),
                [".sub1.a.not.exist == nil"] = check_value("sub1.a.not.exist"),
                [".sub.a.not == nil"] = check_type("sub.a.not", "nil"),
            },
        },
    }

    local initial_t1 = tbl.copy(t1)
    for i, test in ipairs(tests) do
        t1 = test.reset and tbl.copy(initial_t1) or t1
        vim.api.nvim_echo({
            { "Test #" .. tostring(i), "Title" },
        }, true, {})
        if type(test.pre) == "function" then
            test.pre(t1)
        end
        for assert, check_fn in pairs(test.expect) do
            local status, value = check_fn(t1)
            vim.api.nvim_echo({
                { status and "" or "", status and "Normal" or "Error" },
                { "\t" },
                { assert, "Keyword" },
                not status
                    and { "\t" .. tostring(value) }
                    or { "" },
            }, true, {})
        end
    end
end

function M.tests.tables_get(strip, test, subtest)
    local tests = {
        {
            t1 = {},
            paths = {
                "",
                ".",
                "foo",
                "foo.bar.baz",
            },
        },
        {
            t1 = {},
            paths = {
                "",
                "foo.bar.baz",
                "foo.bar.bazz",
                "foo.bar.baz.not_a_table",
                "foo.baz.bar.other",
                "foo.baz.bar.foo",
            },
            defaults = { foo = { bar = { baz = true } } },
        }, {
            t1 = {
                foo = {
                    a = false,
                },
                bar = {
                    foo = {
                        baz = true,
                    },
                    a = true,
                },
                baz = {},
            },
            paths = {
                "",
                "foo",
                "foo.bar",
                "foo.a",
                "foo.bar.baz.b.c",
                "bar.foo",
                "maybe_skipped",
            },
            defaults = {
                foo = {
                    a = "maybe_skipped",
                    bar = {
                        baz = {
                            b = true,
                        },
                    },
                },
                maybe_skipped = {
                    whatever = true,
                },
            },
        }, {
            defaults = {
                overrided_default = {
                    foo = "default",
                },
                empty = {
                    foo = {
                        bar = {
                            baz = {
                                table = {},
                                default = "default",
                            },
                        },
                    },
                },
                same_as_default = {
                    foo = {
                        bar = {
                            baz = {},
                        },
                    },
                },
            },
            t1 = {
                overrided_default = {
                    foo = false,
                },
                empty = {},
                same_as_default = {
                    foo = {
                        bar = {
                            baz = {},
                        },
                    },
                },
            },
            paths = {
                "overrided_default.foo",
                "overrided_default.foo.nope",
                "overrided_default",
                "empty.foo.bar.baz.table.nil",
                "empty.foo.bar.baz.table",
                "empty.foo.bar.baz.default",
                "empty.foo",
                "same_as_default.foo.bar.baz",
                "same_as_default",
            }
        },
    }
    local copy = require("hsv2.utils.tables").copy
    local get = require("hsv2.utils.tables").get
    local t3 = {}
    for i, params in ipairs(tests) do
        if test == nil or test == i then
            local t2 = copy(params.t1)
            vim.api.nvim_echo({ { "Test " .. tostring(i), "Title" } }, true, {})
            vim.pretty_print({ defaults = { params.defaults } })
            for j, path in ipairs(params.paths) do
                if subtest == nil or subtest == j then
                    local t1 = copy(params.t1)
                    vim.pretty_print({ t1 = t1, t2 = t2, t3 = t3 })
                    vim.api.nvim_echo({ {
                        tostring(i) .. "." .. tostring(j) .. " '" .. path .. "':",
                        "Keyword"
                    } }, true, {})
                    local res, status, msg = get(t1, path, params.defaults, strip)
                    if not status then
                        vim.pretty_print({ "t1", error = { res, status, msg } })
                    else
                        vim.pretty_print({ "t1", res = res == nil and "nil" or res, t1 = t1, msg = msg })
                    end
                    res, status, msg = get(t2, path, params.defaults, strip)
                    if not status then
                        vim.pretty_print({ "t2", error = { res, status, msg } })
                    else
                        vim.pretty_print({ "t2", res = res == nil and "nil" or res, t2 = t2, msg = msg })
                    end
                    res, status, msg = get(t3, path, params.defaults, strip)
                    if not status then
                        vim.pretty_print({ "t3", error = { res, status, msg } })
                    else
                        vim.pretty_print({ "t3", res = res == nil and "nil" or res, t3 = t3, msg = msg })
                    end
                end
            end
        end
    end
end

function M.tests.tables_set()
    local set = require("hsv2.utils.tables").set
    local parse = require("hsv2.utils.paths").parse
    local tests = {
        {
            path = "foo",
            value = true,
        },
        {
            path = "foo",
            value = {},
        },
        {
            path = "foo.bar",
            value = true,
        },
        {
            path = "foo.bar",
            value = { baz = true },
        },
        {
            path = "foo.bar.baz.bazz",
            value = true,
        },
        {
            path = "foo.baz.bar.bar.baz",
            value = true,
        },
        {
            path = "foo.bar",
        },
    }
    local t1 = {}
    for i, params in ipairs(tests) do
        local t2 = {}
        local _, res, _, err = pcall(set, t1, parse(params.path), params.value)
        vim.pretty_print({ i, "append", params.path, params.value, res = res, err = err })
        _, res, _, err = pcall(set, t2, parse(params.path), params.value)
        vim.pretty_print({ i, "empty", params.path, params.value, res = res, err = err })
    end
end

function M.fn.run()
end

return M
